/// ReWindowLibrary
/// A C++20 Library for creating and managing windows in a platform-independent manner
/// Developed and Maintained by the boys @ Redacted Software.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file gamepadbutton.h
/// @desc GamepadButton class and enumerations to define standard buttons found on a Gamepad
/// @edit 2024-07-29

#include <string>
#include <J3ML/LinearAlgebra.hpp>

class GamepadButton {
protected:
    std::string mnemonic_btn_code;
public:
    explicit GamepadButton(const std::string& mnemonic) : mnemonic_btn_code(mnemonic) {  }
    [[nodiscard]] std::string GetMnemonicButtonCode() const { return mnemonic_btn_code; }

    /// Compares two GamepadButtons by their mnemonic button codes, not their activation state.
    bool operator ==(const GamepadButton& rhs) const;
};

class GamepadTrigger {
public:

    /// Returns a float value between 0-1 representing how much the trigger has been pushed in by.
    /// (0 being unpressed, 1 being fully pressed)
    float GetActuation() const;

    /// TODO: Might be more appropriate in the Gamepad class representation.
    void SetActuationThreshold(float minimum = 0.01f) const;
};

class GamepadThumbstick
{
public:
    /// Returns a Vector2 value representing the x,y coordinates of the joystick, with 0,0 being the center (at rest).
    /// This vector ranges from length = 0 to length = 1 (i.e. the unit circle).
    [[nodiscard]] Vector2 GetPosition() const;


    /// Sets the deadzone range of the thumbstick.
    /// Deadzone controls how far the stick must be moved before any movement is actually reported.
    /// This is because the thumbstick at-rest will often still report movement.
    /// If gamecode is architected to use the thumbstick position as a direction, without factoring in magnitude, this would cause problems.
    void SetDeadzone(float minimum = 0.01f) const;
};



using J3ML::LinearAlgebra::Vector2;

namespace GamepadButtons {
    static const GamepadButton X {"X"};
    static const GamepadButton Y {"Y"};
    static const GamepadButton A {"A"};
    static const GamepadButton B {"B"};

    static const auto Triangle = Y;
    static const auto Square   = X;
    static const auto Circle   = A;
    static const auto Cross    = B;


    static const GamepadButton LeftBumper {"LB"};
    static const GamepadButton RightBumper {"RB"};
}

namespace GamepadTriggers
{
    static const GamepadTrigger Left;
    static const GamepadTrigger Right;
}