#pragma once
#include <cstdint>
#include <vector>
#include <Event.h>
#include <map>
#include <thread>
//#include <rewindow/types/key.h>
#include <rewindow/types/cursors.h>
#include <rewindow/types/mousebutton.h>
#include <rewindow/types/gamepadbutton.h>
#include <rewindow/types/mouse.h>
#include <rewindow/types/WindowEvents.hpp>
#include <rewindow/types/keyboard.h>
#include <rewindow/types/gamepad.h>
#include <J3ML/LinearAlgebra.hpp>
#include <format>

enum class RWindowFlags: uint8_t {
	IN_FOCUS,
	FULLSCREEN,
	RESIZABLE,
	VSYNC,
	QUIT,
	MAX_FLAG
};

std::string RWindowFlagToStr(RWindowFlags flag);

enum class RenderingAPI: uint8_t {
	OPENGL = 0,
	//Vulkan is unimplemented.
	VULKAN = 1,
};

namespace ReWindow
{
	using J3ML::LinearAlgebra::Vector2;

    /// RWindow is a class implementation of a platform-independent window abstraction.
    /// This library also provides abstractions for user-input devices, and their interaction with the window.
	class RWindow {
	public:
#pragma region Constructors
        /// The default constructor sets a default size and window title.
        RWindow();
        /// Constructs a window by explicitly setting title, width and height.
        RWindow(const std::string& title, int width, int height);
        /// Constructs a window as above with the additional argument of explicitly setting which render API is to be used.
        RWindow(const std::string& title, int width, int height, RenderingAPI renderer);
#pragma endregion

        /// Bindables are provided for hooking into window instances conveniently, without the need to derive and override for simple use cases.
#pragma region Bindable Events

		Event<> OnOpenEvent;
        Event<> OnClosingEvent;
		Event<RWindowEvent> OnFocusLostEvent;
		Event<RWindowEvent> OnFocusGainEvent;
		Event<float> OnRefreshEvent;
		Event<WindowResizeRequestEvent> OnResizeRequestEvent;
        Event<KeyDownEvent> OnKeyDownEvent;
		Event<KeyUpEvent> OnKeyUpEvent;
		Event<MouseMoveEvent> OnMouseMoveEvent;
		Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
		Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
		Event<MouseWheelEvent> OnMouseWheelEvent;

        Event<InputDevice*, InputEvent> OnInputDeviceEvent;
#pragma endregion

        /// These methods can also be overridden in derived classes.
#pragma region Overrides

        /// Called upon the window requesting to open.
		virtual void OnOpen() {}
        /// Called right before the window closes.
        virtual void OnClosing() {}
        /// Called when the window loses focus.
		virtual void OnFocusLost(const RWindowEvent& e) {}
        /// Called when the window gains focus.
		virtual void OnFocusGain(const RWindowEvent& e) {}
        /// Called when the window is 'refreshed', in other words, a render pass is completed.
		virtual void OnRefresh(float elapsed) {}
        /// Called when a resize request has succeeded.
		virtual void OnResizeSuccess() {}
        /// Called when a resize request is sent to the operating system.
		virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
		virtual void OnKeyDown(const KeyDownEvent&) {}
		virtual void OnKeyUp(const KeyUpEvent&) {}
		virtual void OnMouseMove(const MouseMoveEvent&) {}
		virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
		virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
		virtual void OnMouseWheel(const MouseWheelEvent&) {}
#pragma endregion

        Mouse* GetMouse();
        Keyboard* GetKeyboard();

#pragma region Input Helpers
		/// Returns a Vector2 representing mouse coordinates relative to the top-left corner of the window.
		/// This result is cached from the operating-system, and as such may be out-of-date.
		/// @see GetAccurateMouseCoordinates().
		Vector2 GetMouseCoordinates() const;

        int GetMouseWheelPersistent() const { return currentMouse.Wheel;}

#pragma endregion

#pragma region Cursor Modifiers
        void SetCursorStyle(CursorStyle style) const;

        void SetCursorCustomIcon() const;

        void SetCursorLocked();

        void SetCursorCenter();

        void RestoreCursorFromLastCenter(); // Feels nicer for users

        /// Hides the cursor when it's inside of our window. Useful for 3D game camera.
        void SetCursorVisible(bool cursor_enable);

        bool GetCursorVisible();

#pragma endregion

		/// Sets which rendering API is to be used with this window.
		void SetRenderer(RenderingAPI api);

		/// This function instructs the operating system to create the actual window, and give it to us to control.
		/// Calling this function, therefore, creates the real 'window' object on the operating system.
		void Open();

		/// Cleans up and closes the window without destroying the handle.
		void Close();

		void CloseAndReopenInPlace();

		void MessageBox(); // TODO: Must be implemented from scratch as a Motif Window in x11

        /// Sends a request to the operating system to have this window somehow demand attention from the user.
        /// This is implemented per-operating system, and as such no guarantees about behavior can be made.
        void Flash();

		/// Returns whether the window currently has mouse and/or keyboard focus.
		[[nodiscard]] bool IsFocused() const;

        /// Returns whether the window is currently in Fullscreen.
        // TODO: Support Fullscreen, FullscreenWindowed, and Windowed?
		[[nodiscard]] bool IsFullscreen() const;

        /// Returns whether the window can be resized.
		[[nodiscard]] bool IsResizable() const;

        /// Returns whether V-Sync is enabled.
		[[nodiscard]] bool IsVsyncEnabled() const;

        /// Returns whether the window is considered to be alive. Once dead, any logic loop should be terminated, and the cleanup procedure should run.
		[[nodiscard]] bool IsAlive() const;

        /// Returns whether the given key is currently being pressed.
		[[nodiscard]] bool IsKeyDown(Key key) const;

        /// Returns whether the given mouse button is currently being pressed.
		[[nodiscard]] bool IsMouseButtonDown(const MouseButton &button) const;

        /// Sets whether or not to make the window fullscreen.
        /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
		void SetFullscreen(bool fs);

        /// Sets whether or not to make the window resizable.
        /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
		void SetResizable(bool resizable);

        /// Sets whether or not to enable vertical synchronization.
        /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
		void SetVsyncEnabled(bool vsync);

        /// Sets the title of this window.
		void SetTitle(const std::string& title);
		[[nodiscard]] std::string GetTitle() const;

        /// Returns the horizontal length of this window, in pixels.
		[[nodiscard]] int GetWidth() const;
        /// Returns the vertical length of this window, in pixels.
		[[nodiscard]] int GetHeight() const;

        /// A special-case function to change our internal size variable, without triggering event updates.
		void SetSizeWithoutEvent(const Vector2& size); //AAAAAHHHHHHHHH WINDOZE MAKING THINGS DIFFICULT :/ - Redacted.

		void SetLastKnownWindowSize(const Vector2& size);

		// TODO: Josh hates parameter-flags, it's not 1995 :/
		bool GetFlag(RWindowFlags flag) const;
		// TODO: Josh hates parameter-flags, it's not 1995 :/
		void SetFlag(RWindowFlags flag, bool state);

        /// Returns the name of the developer of the user's graphics driver, if it can be determined.
        std::string getGraphicsDriverVendor();

		/// Tells the underlying window manager to destroy this window and drop the handle.
		/// The window, in theory, can not be re-opened after this.
		void DestroyWindow();

		/// Reads events from the operating system, and processes them accordingly.
		/// TODO: Move out of public API, consumers should call Refresh or ideally an update() call.
		void PollEvents();

        /// Updates internal window state, similar to ManagedRefresh, but without accounting for any timekeeping. This is left up to the user.
		void Refresh();

        /// Updates the window and handles timing internally.
        void ManagedRefresh();

        /// Requests the operating system to change the window size.
        /// @param width
        /// @param height
		void SetSize(int width, int height);
        /// Requests the operating system to change the window size.
        /// @param size
		void SetSize(const Vector2& size);

		/// Returns the position of the window's top-left corner relative to the display
		Vector2 GetPos() const;

		/// Returns the known size of the window, in {x,y} pixel measurement.
		Vector2 GetSize() const;

        /// Returns the position of the "renderable area" of the window relative to it's top left corner.
        /// (used to account for the width of the border & title bar).
        Vector2 GetPositionOfRenderableArea() const;

        /// Requests the operating system to move the window to the specified coordinates on the display.
        /// @param x The horizontal screen position to place the window at.
        /// @param y The vertical screen position to place the window at.
		void SetPos(int x, int y);
        /// Requests the operating system to move the window to the specified coordinates on the display.
        /// @param pos A Vector2 representing the x,y coordinates of the desired window destination. Fractional values are ignored.
		void SetPos(const Vector2& pos);


		/// Pull the window to the top, such that it is displayed on top of everything else.
		/// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
		void Raise() const;
		/// Push the window Lower, such that it is effectively hidden behind other windows.
		/// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
		void Lower() const;


		/// Calls OpenGL's SwapBuffers routine.
		/// NOTE: This is only used when the underlying rendering API is set to OpenGL.
		static void GLSwapBuffers();

        /// Returns the current time, represented as a high-resolution std::chrono alias.
        static std::chrono::steady_clock::time_point GetTimestamp();

        /// Computes elapsed time from a start-point and end-point.
        float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end);

        /// Updates internals to account for the latest calculated frame time.
        void UpdateFrameTiming(float frame_time);

        /// Returns the amount of time, in seconds, between the current and last frame.
        /// Technically, no, it returns the elapsed time of the frame, start to finish.
        [[nodiscard]] float GetDeltaTime() const;

        /// Returns the approximate frames-per-second using delta time.
        [[nodiscard]] float GetRefreshRate() const;

        /// Returns the number of frames ran since the windows' creation.
        [[nodiscard]] float GetRefreshCounter() const;
    protected:
        /// Requests the operating system to make the window fullscreen. Saves the previous window size as well.
        void Fullscreen();

        /// Requests the operating system to take the window out of fullscreen mode. Previously saved window size is restored, if possible.
        void RestoreFromFullscreen();

        bool DetectsMouse();
        bool DetectsKeyboard();
        bool DetectsGamepad();

    protected:
#pragma region Data

        std::string title = "Redacted Window";
        int width = 1280;
        int height = 720;

        bool fullscreen_mode = false;
        bool focused = true;
        bool vsync = false;
        bool cursor_visible = true;
        bool open = false;
        bool resizable = true;

        Vector2 lastKnownWindowSize {0, 0};
        bool flags[5];
        std::vector<RWindowEvent> eventLog;

        /// Input Devices;
        Mouse* mouse;
        Keyboard* keyboard;
        Gamepad* gamepad;

        std::vector<InputDevice*> input_devices;


        RenderingAPI renderer = RenderingAPI::OPENGL;

        //Vector2 mouse_coords = {0, 0};
        //Vector2 last_mouse_coords = {0, 0};


        float delta_time = 0.f;
        float refresh_rate = 0.f;
        unsigned int refresh_count = 0;

        // TODO: Implement ringbuffer / circular vector class of some sort.
        float refresh_rate_prev_1 = 0.f;
        float refresh_rate_prev_2 = 0.f;
        float refresh_rate_prev_3 = 0.f;
        float refresh_rate_prev_4 = 0.f;
        float refresh_rate_prev_5 = 0.f;

        float avg_refresh_rate = 0.0f;
#pragma endregion
#pragma region Internals
		/// Returns the most accurate and recent available mouse coordinates.
		/// @note Call this version at most **once** per-frame. It polls the X-Window server and therefore is quite slow.
		/// @see getCursorPos();
		Vector2 GetAccurateMouseCoordinates() const;

	public:
/// These unfortunately *have* to be public because of the poor design of the windows event loop.
#pragma region Event Callers
        /// Executes event handlers for keyboard rele;ase events.
        void processKeyRelease (Key key);
        /// Executes event handlers for keyboard press events.
        void processKeyPress (Key key);
        /// Executes event handlers for window close events.
        /// @note This will be invoked **before** the window-close procedure begins.
		void processOnClose();
        /// Executes event handlers for window open events.
        /// @note This will be invoked **after** the window-open procedure completes.
		void processOnOpen();
		/// Executes event handlers for mouse press events.
        void processMousePress(const MouseButton& btn);
        /// Executes event handlers for mouse release events.
        void processMouseRelease(const MouseButton& btn);
        /// Executes event handlers for window focus events.
        void processFocusIn();
        /// Executes event handlers for window unfocus events.
        void processFocusOut();

        void processMouseMove(Vector2 last_pos, Vector2 new_pos);

        void processMouseWheel(int scrolls);
#pragma endregion
private:
#pragma endregion
    };
}