 #include <rewindow/types/window.h>
#include "rewindow/logger/logger.h"
#include <rewindow/types/mouse.h>

std::string RWindowFlagToStr(RWindowFlags flag) {
    switch (flag) {
        case RWindowFlags::IN_FOCUS: return "IN_FOCUS";
        case RWindowFlags::FULLSCREEN: return "FULLSCREEN";
        case RWindowFlags::RESIZABLE: return "RESIZEABLE";
        case RWindowFlags::VSYNC: return "VSYNC";
        case RWindowFlags::QUIT: return "QUIT";
        case RWindowFlags::MAX_FLAG: return "MAX_FLAG";
        default:
            return "unimplemented flag";
    }
};

using namespace ReWindow;



RWindow::RWindow() {
    title = "ReWindow Application";
    width = 640;
    height = 480;

    if (DetectsKeyboard())
    {
        keyboard = new Keyboard();
        input_devices.push_back(keyboard);
    }

    if (DetectsMouse())
    {
        mouse = new Mouse();
        input_devices.push_back(mouse);
    }

    if (DetectsGamepad())
    {
        gamepad = new Gamepad();
        input_devices.push_back(gamepad);
    }





    //RWindow::singleton = this;
}

RWindow::RWindow(const std::string& title, int width, int height) : flags{false,false,false,false} {
    this->title = title;
    this->width = width;
    this->height = height;
    //RWindow::singleton = this;
}

RWindow::RWindow(const std::string& title, int width, int height, RenderingAPI renderer) :
        title(title), width(width), height(height), renderer(renderer) {}

Vector2 RWindow::GetMouseCoordinates() const {
    return mouse->GetPosition();
}

bool RWindow::GetFlag(RWindowFlags flag) const {
    return flags[(int) flag];
}

bool RWindow::IsAlive() const {
    return true;
}


void RWindow::SetFullscreen(bool fs) {
    if (fs)
        Fullscreen();
    else
        RestoreFromFullscreen();
}

void RWindow::processFocusIn()
{
    RWindowEvent event {};
    OnFocusGain(event);
    OnFocusGainEvent(event);
}

void RWindow::processFocusOut()
{
    RWindowEvent event {};
    OnFocusLost(event);
    OnFocusLostEvent(event);
}

void RWindow::processMousePress(const MouseButton& btn)
{
    mouse->DoButtonPress(btn); // .Set(btn, true);
    auto eventData = MouseButtonDownEvent(btn);
    OnMouseButtonDown(eventData);
    OnMouseButtonDownEvent(eventData);

}

void RWindow::processMouseMove(Vector2 last_pos, Vector2 new_pos)
{
    currentMouse.Position = new_pos;
    auto eventData = MouseMoveEvent(new_pos);
    OnMouseMove(eventData);
    OnMouseMoveEvent(eventData);
}

void RWindow::processMouseRelease(const MouseButton& btn)
{
    currentMouse.Set(btn, false);
    auto eventData = MouseButtonUpEvent(btn);
    OnMouseButtonUp(eventData);
    OnMouseButtonUpEvent(eventData);

}

void RWindow::processKeyRelease(Key key) {
    currentKeyboard.PressedKeys[key] = false;
    auto event = KeyUpEvent(key);
    OnKeyUp(event);
    OnKeyUpEvent(key);
}


std::string RWindow::GetTitle() const {
    return this->title;
}

/*
Vector2 RWindow::GetSize() const
{
    return {this->width, this->height};
}
*/

int RWindow::GetWidth() const
{
    return this->width;
}

int RWindow::GetHeight() const
{
    return this->height;
}

void RWindow::SetSizeWithoutEvent(const Vector2& size) {
    width = size.x;
    height = size.y;
}

void RWindow::SetLastKnownWindowSize(const Vector2& size) {
    lastKnownWindowSize = size;
}

void RWindow::SetRenderer(RenderingAPI api) {
    renderer = api;
}

void RWindow::SetSize(const Vector2& size) {
    this->width = size.x;
    this->height = size.y;
    this->SetSize(size.x, size.y);
}

bool RWindow::IsKeyDown(Key key) const {
    for (const auto& pair : currentKeyboard.PressedKeys)
        if (pair.first == key && pair.second)
            return true;
    return false;
}

bool RWindow::IsMouseButtonDown(const MouseButton &button) const {
    // TODO: Implement MouseButton map
    return currentMouse.IsDown(button);
}

void RWindow::processKeyPress(Key key) {
    currentKeyboard.PressedKeys[key] = true;

    Input::

    auto eventData = KeyDownEvent(key);
    OnKeyDown(eventData);
    OnKeyDownEvent(key);
}

void RWindow::processOnClose()
{
    OnClosing();
    OnClosingEvent();
}

void RWindow::processOnOpen()
{
    OnOpen();
    OnOpenEvent();
}

void RWindow::ManagedRefresh()
{
     auto begin = GetTimestamp();
     Refresh();
     auto end = GetTimestamp();

     float dt = ComputeElapsedFrameTimeSeconds(begin, end);

     UpdateFrameTiming(dt);
}

void RWindow::Refresh() {
    PollEvents();
    OnRefresh(delta_time);

    // Only call once and cache the result.
    currentMouse.Position = GetAccurateMouseCoordinates();

    /// TODO: Implement optional minimum epsilon to trigger a Mouse Update.
    if (currentMouse.Position != previousMouse.Position) {
        processMouseMove(previousMouse.Position, currentMouse.Position);
        previousMouse.Position = currentMouse.Position;
    }

    detail::UpdateMouse(currentMouse);

}

float RWindow::ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end) {
    auto frame_time = end - start;
    unsigned long int frame_time_us = std::chrono::duration_cast<std::chrono::microseconds>(frame_time).count();
    float frame_time_s = frame_time_us / (1000.f * 1000.f);
    return frame_time_s;
}

 std::chrono::steady_clock::time_point RWindow::GetTimestamp() {
    return std::chrono::steady_clock::now();
}

void RWindow::UpdateFrameTiming(float frame_time) {
    delta_time = frame_time;
    refresh_rate = 1.f / delta_time;
    refresh_rate_prev_5 = refresh_rate_prev_4;
    refresh_rate_prev_4 = refresh_rate_prev_3;
    refresh_rate_prev_3 = refresh_rate_prev_2;
    refresh_rate_prev_2 = refresh_rate_prev_1;
    refresh_rate_prev_1 = refresh_rate;
    avg_refresh_rate = (refresh_rate_prev_1 + refresh_rate_prev_2 + refresh_rate_prev_3 + refresh_rate_prev_4 + refresh_rate_prev_5) / 5.f;
    refresh_count++;
}


void RWindow::processMouseWheel(int scrolls)
{
    currentMouse.Wheel += scrolls;
    auto ev = MouseWheelEvent(scrolls);
    OnMouseWheel(ev);
    OnMouseWheelEvent(ev);

    previousMouse.Wheel = currentMouse.Wheel;
}


void RWindow::Close() {
    /// TODO: Implement closing the window without destroying the handle.
    processOnClose();
}

bool RWindow::IsResizable() const {
    return resizable;
}

bool RWindow::IsFullscreen() const {
    return fullscreen_mode;
}

bool RWindow::IsFocused() const {
    return focused;
}

 bool RWindow::IsVsyncEnabled() const {
     return vsync;
 }

 float RWindow::GetDeltaTime() const { return delta_time; }

 float RWindow::GetRefreshRate() const { return refresh_rate; }

 float RWindow::GetRefreshCounter() const { return refresh_count; }

